#region [ License information          ]
/* ************************************************************
 *
 * Copyright (c) Bjrn Bouet Smith, 2010, 2011
 *
 * This source code is subject to terms and conditions of 
 * Microsoft Public License (Ms-PL).
 * 
 * A copy of the license can be found in the license.txt
 * file at the root of this distribution. If you can not 
 * locate the License, please send an email to bjornsmith@gmail.com
 * 
 * By using this source code in any fashion, you are 
 * agreeing to be bound by the terms of the Microsoft 
 * Permissive License.
 *
 * You must not remove this notice, or any other, from this
 * software.
 *
 * ************************************************************/
#endregion

using System;
using System.Net;

namespace Smith.AMC.Network
{
    /// <summary>
    /// Represents a memcached server node, that is part of the mecached server cluster.
    /// </summary>
    public interface IMemcachedServerNode : IDisposable
    {
        /// <summary>
        /// Gets or sets a value indicating whether this instance is dead.
        /// To be used to remove a server node from the server pool.
        /// </summary>
        /// <value><c>true</c> if this instance is dead; otherwise, <c>false</c>.</value>
        bool IsDead
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the failed count, i.e. the number of times
        /// a socket have been reported as dead for this server node.
        /// 
        /// This is being used to control whether or not the node should be marked as dead
        /// </summary>
        /// <value>The failed count.</value>
        int FailedCount
        {
            get;
            set;
        }

        /// <summary>
        /// Gets the server address this server node is listening on.
        /// </summary>
        IPAddress ServerAddress
        {
            get;
        }

        /// <summary>
        /// Gets or sets the server ID.
        /// </summary>
        int ServerID
        {
            get;
            set;
        }

        /// <summary>
        /// Gets the port number this server node is listening on
        /// </summary>
        /// <value>The port number.</value>
        int PortNumber
        {
            get;
        }

        /// <summary>
        /// Gets the health check socket that is being used to determine whether or not this instance is still alive
        /// when too many clients have reported errors on the memcached node.
        /// </summary>
        /// <value>The health check socket.</value>
        IClientSocket HealthCheckSocket
        {
            get;
        }

        /// <summary>
        /// Aquires a client socket to be used with operations against this instance of the mecached node.
        /// </summary>
        /// <param name="timeout">The timeout time for obtaining a socket.</param>
        /// <returns>The socket, ready to be used.</returns>
        /// <exception cref="TimeoutException">If the timeout period expired</exception>
        IClientSocket AquireSocket(TimeSpan timeout);

        /// <summary>
        /// Releases the socket back into the socket pool, ready for another thread to pick it up.
        /// </summary>
        /// <param name="socket">The socket to release back.</param>
        void ReleaseSocket(IClientSocket socket);

        /// <summary>
        /// Renews the failed socket by removing the given socket from the queue and recreating it.
        /// Will increment the failed socket count for this server node.
        /// </summary>
        /// <param name="socket">The socket.</param>
        void RenewFailedSocket(IClientSocket socket);
    }
}